import warnings

import numpy as np


def _get_interception_at_xaxis(xlist, ylist):
    domain = np.std(xlist) * np.arange(-1, 2, 2) + np.mean(xlist)
    p = np.polynomial.Polynomial.fit(xlist, ylist, 1, domain=domain)
    XIntercept = p.roots()[0]  # 6.336226562500000e+03
    return XIntercept, p


def get_xy_line_cross(line_eq):
    xLineCross = line_eq.roots()[0]
    yLineCross = line_eq(xLineCross)
    return xLineCross,yLineCross


def _get_tent_coord(xLeft, yLeft,xRight, yRight):
    ## Let get the tent point
    left_line = np.polynomial.Polynomial.fit(xLeft, yLeft, 1, domain=[-1, 1])
    right_line = np.polynomial.Polynomial.fit(xRight, yRight, 1, domain=[-1, 1])
    xIntersect = (left_line - right_line).roots()[0]  # x0 6.192071289062500e+03
    yIntersect = left_line(xIntersect)  # y0=9.931675910949707

    ## Important for plotting
    xLineCross_l,yLineCross_l=get_xy_line_cross(left_line)


    xLineCross_r,yLineCross_r=get_xy_line_cross(right_line)


    return xIntersect,yIntersect,xLineCross_l,yLineCross_l,xLineCross_r,yLineCross_r

def get_line_intersection_slope(xIntersect,yIntersect,leftXIntercept,rightXIntercept):
    leftSlope = yIntersect / (xIntersect - leftXIntercept)  # 0.36513907
    rightSlope = yIntersect / (xIntersect - rightXIntercept)  # -0.068895683
    return leftSlope,rightSlope

def get_average_velocity(pLeft,pRight,xLeft,xRight):
    averLeftVelocity = pLeft.coef[1] / np.std(xLeft)  # 0.36513701
    averRightVelocity = pRight.coef[1] / np.std(xRight)  # -0.068895057

    return averLeftVelocity,averRightVelocity
def  line_corrcoef(p,xlist,ylist):

    ffit = np.polyval(p.coef, xlist)

    R2 = np.corrcoef(ylist.flatten(), ffit.flatten())[0, 1]
    return R2


def lines_intersection(xRight=None, xLeft=None, yRight=None, yLeft=None, dic_type=True):
    """
    Source:
    https://stackoverflow.com/q/68182079/6446053

    I think tent is the top most point
    leftSlope
    The slope of the left tent line or NaN if the tent line doesn’t exist.

    rightSlope
    Slope of the right blink tent line or NaN if the tent line doesn’t exist.

    averLeftVelocity
    The velocity as estimated by the left tent line.

    averRightVelocity
    The velocity as estimated by the right tent line.

    leftR2
    The correlation of the left tent line with the 80% blink upstroke. [A numeric value or NaN if the tent line doesn’t exist.]

    rightR2
    The correlation of the right tent line with the 80% blink downstroke. [A numeric value or NaN if the tent line doesn’t exist.]

    xIntersect
    The x-coordinate of the intersection of the left and right tent line with the x-axis.
    [A numeric value or NaN if the tent line doesn’t exist.]

    yIntersect
    The y-coordinate of the intersection of the left and right tent line with the x-axis.
    [A numeric value or NaN if the tent line doesn’t exist.]

    leftXIntercept
    An integer giving the frame number of the intersection of the left tent line with the x-axis (or NaN if the tent line doesn’t exist).

    rightXIntercept
    An integer giving the frame number of the intersection of the right tent line with the x-axis (or NaN if the tent line doesn’t exist).


    """

    rightXIntercept, pRight = _get_interception_at_xaxis(xRight, yRight)
    leftXIntercept, pLeft = _get_interception_at_xaxis(xLeft, yLeft)



    # ## Let get the tent point
    xIntersect,yIntersect,xLineCross_l,yLineCross_l,xLineCross_r,yLineCross_r=_get_tent_coord(xLeft, yLeft,xRight, yRight)

    leftSlope,rightSlope=get_line_intersection_slope(xIntersect,yIntersect,leftXIntercept,rightXIntercept)

    averLeftVelocity,averRightVelocity=get_average_velocity(pLeft,pRight,xLeft,xRight)

    rightR2,leftR2 =line_corrcoef(pRight,xRight,yRight),line_corrcoef(pLeft,xLeft,yLeft)


    # from eeg_blinks.viz.viz_sanity import viz_line_intersection
    # viz_line_intersection(candidateSignal,xRight,xLeft,leftXIntercept,rightXIntercept,
    #                       xIntersect,yIntersect)

    if dic_type:
        d = dict(leftSlope=leftSlope, rightSlope=rightSlope, averLeftVelocity=averLeftVelocity,
                 averRightVelocity=averRightVelocity, rightR2=rightR2, leftR2=leftR2)
        return d
    else:
        return leftSlope, rightSlope, averLeftVelocity, averRightVelocity, \
               rightR2, leftR2, xIntersect, yIntersect, leftXIntercept, rightXIntercept, \
               xLineCross_l, yLineCross_l, xLineCross_r, yLineCross_r


def left_right_zero_crossing(candidateSignal, maxFrame, outerStarts, outerEnds):
    ### Latest as of 29 April 2022 which is more efficient
    theRange = np.arange(int(outerStarts), int(maxFrame))
    sInd_leftZero = np.flatnonzero(candidateSignal[theRange] < 0)

    if (sInd_leftZero.size != 0):
        leftZero = theRange[sInd_leftZero[-1]]

    else:
        extreme_outerStartss = np.arange(0, maxFrame)
        extreme_outerStartss = extreme_outerStartss.astype(int)
        sInd_rightZero_ex = np.flatnonzero(candidateSignal[extreme_outerStartss] < 0)[-1]
        leftZero = extreme_outerStartss[sInd_rightZero_ex]

    theRange = np.arange(int(maxFrame), int(outerEnds))
    sInd_rightZero = np.flatnonzero(candidateSignal[theRange] < 0)

    if (sInd_rightZero.size != 0):
        rightZero = theRange[sInd_rightZero[0]]
    else:
        """
        We take extreme remedy by extending the outerEnds to the maximum
        """
        extreme_outerEns = np.arange(maxFrame, candidateSignal.shape)
        extreme_outerEns = extreme_outerEns.astype(int)
        sInd_rightZero_ex_s = np.flatnonzero(candidateSignal[extreme_outerEns] < 0)

        if (sInd_rightZero_ex_s.size != 0):
            # This usually happen for end of signal
            sInd_rightZero_ex = sInd_rightZero_ex_s[0]
            rightZero = extreme_outerEns[sInd_rightZero_ex]
        else:
            return leftZero, None

    if leftZero > maxFrame:
        raise ValueError('something is not right')

    if maxFrame > rightZero:
        raise ValueError('something is not right')

    # from eeg_blinks.viz.viz_sanity import _viz_sanity_zero_crossing
    # _viz_sanity_zero_crossing(leftZero, rightZero, maxFrame, candidateSignal, 'Zero_Crossing')

    return leftZero, rightZero


def get_up_down_stroke(maxFrame, leftZero, rightZero):
    upStroke = np.arange(leftZero, maxFrame)
    downStroke = np.arange(maxFrame, rightZero)
    return upStroke, downStroke


def _maxPosVelFrame(blinkVelocity, maxFrame, leftZero, rightZero):
    maxFrame, leftZero, rightZero = int(maxFrame), int(leftZero), int(rightZero)
    upStroke, downStroke = get_up_down_stroke(maxFrame, leftZero, rightZero)
    maxPosVelFrame = np.argmax(blinkVelocity[upStroke])
    maxPosVelFrame = maxPosVelFrame + upStroke[0]

    if len(blinkVelocity[downStroke])>0:
        maxNegVelFrame = np.argmin(blinkVelocity[downStroke])
        maxNegVelFrame = maxNegVelFrame + downStroke[0]
    else:
        warnings.warn('Force nan but require further invesitigation why happen like this')
        maxNegVelFrame=np.nan


    return maxPosVelFrame, maxNegVelFrame


def _get_left_base(blinkVelocity, leftOuter, maxPosVelFrame):
    leftOuter, maxPosVelFrame = int(leftOuter), int(maxPosVelFrame)
    leftBase = np.arange(leftOuter, maxPosVelFrame)

    leftBaseVelocity = np.flip(blinkVelocity[leftBase])

    leftBaseIndex = np.argmax(leftBaseVelocity <= 0)

    leftBase = maxPosVelFrame - leftBaseIndex

    return leftBase


def _get_right_base(candidateSignal, blinkVelocity, rightOuter, maxNegVelFrame):
    # Start Line 102 Matlab
    rightOuter, maxNegVelFrame = int(rightOuter), int(maxNegVelFrame)
    a_tend = np.minimum(rightOuter, candidateSignal.size)

    if maxNegVelFrame > a_tend:
        # warnings.warn(
        #     'Failed to fit blink %s but due to MaxNegVelFrame: %s larger than a_tend: %s .For now I will skip this file'
        #     % (number, maxNegVelFrame, a_tend))
        return None

    rightBase = np.arange(maxNegVelFrame, a_tend)  # Line 102 matlab

    # hh=blinkVelocity.size
    # nn=np.max(rightBase)

    if rightBase.size == 0:
        return None

    if np.max(rightBase) >= blinkVelocity.size:
        # For some reason, the original rightBase has index value greate than blinkVelocity which cause index error.
        # To address this issue, we remove some value
        rightBase = rightBase[:-1]
        if np.max(rightBase) >= blinkVelocity.size:
            raise ValueError('Please strategise how to address this')

    rightBaseVelocity = blinkVelocity[rightBase]  #

    '''
    if rightBaseIndex.size == 0:  # Line 108 Matlab
        rightBaseIndex = 0
    '''
    rightBaseIndex = np.argmax(rightBaseVelocity >= 0)
    # try:
    #     rightBaseIndex = np.argwhere(rightBaseVelocity >= 0)[0]
    # except IndexError:  # Line 108 Matlab
    #     rightBaseIndex = 0

    rightBase = maxNegVelFrame + rightBaseIndex
    # rightBase_t = rightBase / srate

    return rightBase


def _get_half_height(candidateSignal, maxFrame, leftZero, rightZero, leftBase, rightOuter):
    ####
    """
    leftBaseHalfHeight
    The coordinate of the signal halfway (in height) between the blink maximum and the left base value. [A positive numeric value.]

    rightBaseHalfHeight
    The coordinate of the signal halfway (in height) between the blink maximum and the right base value.
    [A positive numeric value.]
    """

    maxFrame, leftZero, rightZero, leftBase, rightOuter = int(maxFrame), int(leftZero), int(rightZero), int(
        leftBase), int(rightOuter)

    ryy = candidateSignal[maxFrame] - candidateSignal[leftBase]
    blinkHalfHeight = candidateSignal[maxFrame] - (0.5 * (ryy))

    leftHalfBase = np.arange(leftBase, maxFrame)
    leftBaseHalfHeight = leftBase + np.argmax(candidateSignal[leftHalfBase] >= blinkHalfHeight)

    warnings.warn(
        'Need to double check this line:To confirm whether it is correct to used rightOuter instead of rightBase?')
    ## WIP : To confirm whether it is correct to used rightOuter instead of rightBase?
    rightHalfBase = np.arange(maxFrame, rightOuter)
    rightBaseHalfHeight = np.minimum(rightOuter,
                                     np.argmax(candidateSignal[rightHalfBase] <= blinkHalfHeight) + maxFrame)

    # Compute the left and right half-height frames from zero
    leftHalfBase = np.arange(leftZero, maxFrame)
    blinkHalfHeight = 0.5 * candidateSignal[maxFrame]  # with_val 4.3747134

    """
    leftZeroHalfHeight
    The coordinate of the signal halfway (in height) between the blink maximum and the left zero value.
    """
    leftZeroHalfHeight = np.argmax(candidateSignal[leftHalfBase] >= blinkHalfHeight) + leftZero

    rightHalfBase = np.arange(maxFrame, rightZero)

    """
    rightZeroHalfHeight
    The coordinate of the signal halfway (in height) between the blink maximum and the right zero value.
    """
    rightZeroHalfHeight = np.minimum(rightOuter, maxFrame +
                                     np.argmax(candidateSignal[rightHalfBase] <= blinkHalfHeight))

    return leftZeroHalfHeight, rightZeroHalfHeight, leftBaseHalfHeight, rightBaseHalfHeight


def compute_fit_range(candidateSignal, maxFrame, leftZero, rightZero, baseFraction, top_bottom=None):
    maxFrame, leftZero, rightZero = int(maxFrame), int(leftZero), int(rightZero)
    # Compute fit ranges
    blinkHeight = candidateSignal[maxFrame] - candidateSignal[leftZero]  # ?? 8.8286028

    blinkTop = candidateSignal[maxFrame] - baseFraction * blinkHeight  # ?? 7.8665667
    blinkBottom = candidateSignal[leftZero] + baseFraction * blinkHeight  # ?? 0.80368418

    blinkRange_l = np.arange(leftZero, maxFrame + 1, dtype=int)


    blinkTopPoint_l = np.argmin(candidateSignal[blinkRange_l] < blinkTop)
    blinkTopPoint_l_X = blinkRange_l[blinkTopPoint_l]
    blinkTopPoint_l_Y = candidateSignal[blinkTopPoint_l_X]

    blinkBottomPoint_l = np.argmax(candidateSignal[blinkRange_l] > blinkBottom)
    blinkBottomPoint_l_X = blinkRange_l[blinkBottomPoint_l]
    blinkBottomPoint_l_Y = candidateSignal[blinkBottomPoint_l_X]


    leftRange = [blinkRange_l[blinkBottomPoint_l], blinkRange_l[blinkTopPoint_l]]


    blinkRange_r = np.arange(maxFrame, rightZero + 1, dtype=int)
    blinkTopPoint_r = np.argmax(candidateSignal[blinkRange_r] < blinkTop)
    blinkTopPoint_r_X = blinkRange_r[blinkTopPoint_r]
    blinkTopPoint_r_Y = candidateSignal[blinkTopPoint_r_X]

    blinkBottomPoint_r = np.argmin(candidateSignal[blinkRange_r] > blinkBottom)
    blinkBottomPoint_r_X = blinkRange_r[blinkBottomPoint_r]
    blinkBottomPoint_r_Y = candidateSignal[blinkBottomPoint_r_X]

    rightRange = [blinkRange_r[blinkTopPoint_r], blinkRange_r[blinkBottomPoint_r]]
    # use this to visualise
    # from eeg_blinks.viz.viz_sanity import viz_blink_top_buttom_point
    # viz_blink_top_buttom_point(candidateSignal,blinkRange,blinkTop,blinkBottom,maxFrame,rightZero,leftRange,rightRange)

    xLeft = np.arange(leftRange[0], leftRange[1] + 1, dtype=int)  # THe +1 to ensure we include the last frame
    xRight = np.arange(rightRange[0], rightRange[1] + 1, dtype=int)



    if blinkBottomPoint_l_X == blinkTopPoint_l_X:
        warnings.warn('same value for left top_blink and left bottom_blink')

    if blinkBottomPoint_r_X == blinkTopPoint_r_X:
        warnings.warn('same value for right top_blink and right bottom_blink')

    if xLeft.size == 0:
        xLeft = np.nan

    if xRight.size == 0:
        xRight = np.nan

    if top_bottom is None:
        warnings.warn('To modify this so that all function return the top_bottom point')
        return xLeft, xRight, leftRange, rightRange
    else:
        return xLeft, xRight, leftRange, rightRange, \
               blinkBottomPoint_l_Y,blinkBottomPoint_l_X,blinkTopPoint_l_Y,blinkTopPoint_l_X,\
               blinkBottomPoint_r_X,blinkBottomPoint_r_Y,blinkTopPoint_r_X,blinkTopPoint_r_Y




def get_zero_crossing_pd(number, maxFrame, maxValue, leftOuter, rightOuter, candidateSignal, blinkVelocity,
                         baseFraction):
    """

    number
    The number of the potential blink within the corresponding blinks structure. [A positive numeric value.]

    maxFrame
    The frame number of the first maximum amplitude of this blink. [A positive numeric value.]

    maxValue
    A numeric value giving the maximum value of the blink.

    leftOuter
    The frame number of the left outer reach of the blink
    (defined as the largest of 1 or the frame of the previous blink maximum).

    rightOuter
    The frame number of the right outer reach of the blink (defined as the smallest of the last
    frame or the frame of the next blink maximum).

    leftZero
    The frame number of the left zero crossing of the blink. [A positive numeric value.]

    rightZero
    The frame number of the right zero crossing of the blink. [A positive numeric value.]
    """


    leftZero, rightZero = left_right_zero_crossing(candidateSignal, maxFrame,
                                                   leftOuter, rightOuter)

    if (leftZero is None) or (rightZero is None):
        # This usually happen for end of signal
        return None
    ## Compute the place of maximum positive and negative velocities
    ## START FROM LINE 87 MATLAB

    """
    
    upStroke is the interval between leftZero and maxFrame, 
    downStroke is the interval between maxFrame and rightZero`.
    """

    maxPosVelFrame, maxNegVelFrame = _maxPosVelFrame(blinkVelocity, maxFrame, leftZero, rightZero)

    assert maxNegVelFrame.size == 1

    # Compute the left and right base frames

    if leftOuter > maxPosVelFrame:
        warnings.warn(
            'Failed to fit blink %s due to leftOuter :%s  larger than maxPosVelFrame : %s. For now I will skip this file'
            % (number, leftOuter, maxPosVelFrame))
        return

    """
    
    leftBase
    The frame number of the left local minimum of the blink. [A positive numeric value.]
    
    rightBase
    The frame number of the right local minimum of the blink. [A positive numeric value.]
    """

    leftBase = _get_left_base(blinkVelocity, leftOuter, maxPosVelFrame)

    rightBase = _get_right_base(candidateSignal, blinkVelocity, rightOuter, maxNegVelFrame)

    leftZeroHalfHeight, rightZeroHalfHeight, leftBaseHalfHeight, rightBaseHalfHeight = _get_half_height(candidateSignal,
                                                                                                        maxFrame,
                                                                                                        leftZero,
                                                                                                        rightZero,
                                                                                                        leftBase,
                                                                                                        rightOuter)

    """
    I am just curios why they dont simply define xLeft  and xRight simply as
    
    xLeft=np.arange(leftZero, maxFrame)
    xRight=np.arange(maxFrame,rightZero)
    """
    xLeft, xRight, leftRange, rightRange = compute_fit_range(candidateSignal, maxFrame, leftZero, rightZero,
                                                             baseFraction)

    if (xLeft.size <= 1) or (xRight.size <= 1):
        # To avoid ValueError: On entry to DLASCL parameter number 4 had an illegal value
        return None

    # Below and above for types
    if (xLeft.size != 0) and (xRight.size != 0):
        # There should be multiple x,y coordinate to perform the math operation

        ds_lines = lines_intersection(xRight=xRight, xLeft=xLeft,
                                      yRight=candidateSignal[xRight], yLeft=candidateSignal[xLeft])

        data_blink = dict(leftRange=leftRange, rightRange=rightRange,
                          rightBaseHalfHeight=rightBaseHalfHeight, leftBaseHalfHeight=leftBaseHalfHeight,
                          leftZero=leftZero, rightZero=rightZero,
                          leftBase=leftBase, rightBase=rightBase,
                          rightZeroHalfHeight=rightZeroHalfHeight, leftZeroHalfHeight=leftZeroHalfHeight,
                          rightOuter=rightOuter, leftOuter=leftOuter,
                          maxFrame=maxFrame, number=number, maxValues=maxValue)

        data_blink.update(ds_lines)

        return data_blink

    else:

        if (xLeft.size == 0) and (xRight.size == 0):
            warnings.warn('Failed to fit blink %s due to xLeft or xRight less than 1: xleft size:%s and xRIght size: %s'
                          % (number, len(xLeft), len(xRight)))
            return None

        else:
            warnings.warn(
                'Failed to fit blink %s but due to imbalance xLeft or xRight less than 1: xleft size:%s and xRIght size: %s. '
                'For now I will skip this file'
                % (number, len(xLeft), len(xRight)))

            return None





import logging

from pyblinkers.utilities.zero_crossing import (_get_half_height,
                                                compute_fit_range, lines_intersection)
from pyblinkers.vislab.base_left_right import create_left_right_base_vislab

logging.getLogger().setLevel(logging.INFO)

class FitBlinks:

    def __init__(self, data=None, df=None):
        self.data = data
        self.df = df
        self.frame_blinks=[]
        self.baseFraction = 0.1
        self.cols_half_height = ['leftZeroHalfHeight', 'rightZeroHalfHeight', 'leftBaseHalfHeight', 'rightBaseHalfHeight']
        self.cols_fit_range = ['xLeft', 'xRight', 'leftRange', 'rightRange',
                               'blinkBottomPoint_l_Y', 'blinkBottomPoint_l_X', 'blinkTopPoint_l_Y', 'blinkTopPoint_l_X',
                               'blinkBottomPoint_r_X', 'blinkBottomPoint_r_Y', 'blinkTopPoint_r_X', 'blinkTopPoint_r_Y']
        self.cols_lines_intesection = ['leftSlope', 'rightSlope', 'averLeftVelocity', 'averRightVelocity',
                                       'rightR2', 'leftR2', 'xIntersect', 'yIntersect', 'leftXIntercept',
                                       'rightXIntercept', 'xLineCross_l', 'yLineCross_l', 'xLineCross_r', 'yLineCross_r']
        self.fit()


    def fit(self):
        self.frame_blinks = create_left_right_base_vislab(self.data, self.df)
        self.frame_blinks[self.cols_half_height] = self.frame_blinks.apply(lambda x: _get_half_height(self.data, x['maxFrames'], x['leftZero'], x['rightZero'],
                                                                                       x['leftBase'], x['outerEnds']), axis=1,
                                                       result_type="expand")
        self.frame_blinks[self.cols_fit_range] = self.frame_blinks.apply(lambda x: compute_fit_range(self.data, x['maxFrames'], x['leftZero'], x['rightZero'],
                                                                                      self.baseFraction, top_bottom=True), axis=1,
                                                     result_type="expand")
        self.frame_blinks = self.frame_blinks.dropna()
        self.frame_blinks['nsize_xLeft'] = self.frame_blinks.apply(lambda x: x['xLeft'].size, axis=1)
        self.frame_blinks['nsize_xRight'] = self.frame_blinks.apply(lambda x: x['xRight'].size, axis=1)
        self.frame_blinks = self.frame_blinks[~(self.frame_blinks['nsize_xLeft'] <= 1) & ~(self.frame_blinks['nsize_xRight'] <= 1)]
        self.frame_blinks.reset_index(drop=True, inplace=True)
        self.frame_blinks[self.cols_lines_intesection] = self.frame_blinks.apply(lambda x: lines_intersection(xRight=x['xRight'], xLeft=x['xLeft'],
                                                                                          yRight=self.data[x['xRight']], yLeft=self.data[x['xLeft']],
                                                                                          dic_type=False), axis=1, result_type="expand")



import warnings

import numpy as np

from pyblinkers.utilities.extractBlinkProperties import _goodblink_based_corr_median_std


class PrepareSelection:
    def __init__(self, signal=None, df=None, params=None, ch=None):
        self.signal = signal
        self.df = df
        self.params = params
        self.ch = ch



    def get_param_for_selection(self):
        self.df['rightR2'] = self.df['rightR2'].abs()
        blink_mask = np.zeros(self.signal.size, dtype=bool)

        for left_zero_y, right_zero_x in zip(self.df['leftZero'] - 1, self.df['rightZero']):
            blink_mask[int(left_zero_y):int(right_zero_x)] = True

        outside_blink = np.logical_and(self.signal > 0, ~blink_mask)
        inside_blink = np.logical_and(self.signal > 0, blink_mask)

        blink_amp_ratio = np.mean(self.signal[inside_blink]) / np.mean(self.signal[outside_blink])



        R2_top, R3_top, best_median, best_robust_std = _goodblink_based_corr_median_std(self.df, self.params['correlationThresholdTop'])
        R2_bot, R3_bot, worst_median, worst_robust_std = _goodblink_based_corr_median_std(self.df,self.params['correlationThresholdBottom'])
        true_top = R2_top.values & R3_top.values
        true_bot = R2_bot.values & R3_bot.values
        good_values = self.df.loc[true_bot, 'maxValues']

        all_values = self.df['maxValues']
        cutoff = (best_median * worst_robust_std + worst_median * best_robust_std) / (best_robust_std + worst_robust_std)


        mask = np.logical_and(all_values <= best_median + 2 * best_robust_std, all_values >= best_median - 2 * best_robust_std)
        all_X = np.count_nonzero(mask)

        if all_X != 0:
            good_ratio = np.sum(np.logical_and(good_values <= best_median + 2 * best_robust_std, good_values >= best_median - 2 * best_robust_std)) / all_X
        else:
            good_ratio = np.nan

        number_good_blinks = true_bot.sum().item()
        all_data = [self.ch, blink_amp_ratio, best_median, best_robust_std, cutoff, good_ratio, number_good_blinks, self.df]
        header_eb_label = ['ch', 'blinkAmpRatio', 'bestMedian', 'bestRobustStd', 'cutoff', 'goodRatio',
                           'numberGoodBlinks']
        data_blink = dict(zip(header_eb_label, all_data))
        return data_blink




class ChannelSelection:
    '''

    Reduce the number of candidate signals based on these steps
    1) Reduce the number of candidate signals based on the blink amp ratios:
        -params ['params_blinkAmpRange_1'],params ['params_blinkAmpRange_2']
    2) Find the ones that meet the minimum good blink threshold
        -params ['params_goodRatioThreshold']
    3) See if any candidates meet the good blink ratio criteria

    4) Pick the one with the maximum number of good blinks

    '''
    def __init__(self, df, params):
        self.df = df
        self.params = params
        self.nbest = 5
        self.nbest_force = 2
        self.params_blinkAmpRange_1 = params['params_blinkAmpRange_1']
        self.params_blinkAmpRange_2 = params['params_blinkAmpRange_2']
        self.params_goodRatioThreshold = params['params_goodRatioThreshold']
        self.params_minGoodBlinks = params['params_minGoodBlinks']
        self.reduce_candidate_signals()
        self.pick_best_channel()
        self.df.reset_index(drop=True, inplace=True)

    def reduce_candidate_signals(self):
        self.df['con_blinkAmpRange'] = np.where((self.df.blinkAmpRatio >= self.params_blinkAmpRange_1) &
                                                (self.df.blinkAmpRatio <= self.params_blinkAmpRange_2),
                                                True, False)

        self.df['con_GoodBlinks'] = np.where((self.df.numberGoodBlinks > self.params_minGoodBlinks),
                                             True, False)

        self.df['con_GoodRatio'] = np.where((self.df.goodRatio > self.params_goodRatioThreshold),
                                            True, False)
        self.df.sort_values(['goodRatio', 'numberGoodBlinks'], ascending=[False, False], inplace=True)

    def pick_best_channel(self):

        # step 1
        nblinkAmpRange = self.df['con_blinkAmpRange'].sum()
        ncon_GoodRatio = self.df['con_GoodRatio'].sum()

        if nblinkAmpRange == 0:
            warnings.warn(f'Blink amplitude ratio too low than the predeterimined therehold-- may be noise.')
            if ncon_GoodRatio != 0:
                self.df = self.df[self.df['con_GoodRatio'] == True]
                self.df = self.df.head(self.nbest_force)
            else:
                self.df = self.df.head(self.nbest_force)
            return self.df

        # Step 2
        # Now see if any candidates meet the good blink ratio criteria

        if ncon_GoodRatio == 0:
            if nblinkAmpRange != 0:
                self.df = self.df[self.df['con_blinkAmpRange'] == True]
                self.df = self.df.head(self.nbest_force)
                return self.df
            else:
                return self.df.head(self.nbest_force)

        # If we fulfill the con_GoodRatio,ncon_GoodRatio
        df1 = self.df[(self.df['con_GoodRatio'] == True) & (self.df['con_blinkAmpRange'] == True)]
        if df1.empty:
            """
            Most probably we have
            con_blinkAmpRange   con_goodBLinks  con_GoodRatio
            FALSE               TRUE            TRUE
            TRUE                TRUE            FALSE
            """
            self.df = self.df[(self.df['con_GoodRatio'] == True) & (self.df['con_GoodBlinks'] == True)]
            return self.df
        else:
            """
            Ideal case
            """
            self.df = self.df[(self.df['con_GoodRatio'] == True) & (self.df['con_blinkAmpRange'] == True) & (self.df['con_GoodBlinks'] == True)]

            return self.df.head(self.nbest)






# import hickle as hkl
import numpy as np
import matplotlib.pyplot as plt

def viz_line_intersection(candidateSignal,xRight,xLeft,leftXIntercept,rightXIntercept,
                          xIntersect,yIntersect):
    #### make_Plot


    srate=100
    npad=20
    idx_t=(np.arange ( np.min(xLeft)-npad, np.max(xRight) +npad))
    bTrace= candidateSignal[idx_t]
    t=idx_t
    #
    plt.plot(t, bTrace, linewidth=10,alpha=0.5)


    plt.plot([leftXIntercept[0],xIntersect[0]], [0,yIntersect[0]],'--')


    plt.plot([xIntersect[0],rightXIntercept[0],], [yIntersect[0],0],'--')
    # plt.show()


    z=[leftXIntercept,rightXIntercept]
    y=[0,0]

    n =  [f'XIntercept_{float("{0:.2f}".format(i[0]))}' for i in z]
    plt.scatter(z,y)

    for label, x, y in zip(n, z,y):
        plt.annotate(
            label,
            xy=(x, y), xytext=(-20, 20),
            textcoords='offset points', ha='right', va='bottom',
            bbox=dict(boxstyle='round,pad=0.5', fc='yellow', alpha=0.5),
            arrowprops=dict(arrowstyle = '->', connectionstyle='arc3,rad=0'))


    plt.scatter([xIntersect],[yIntersect])

    for label, x, y in zip([f'xy_intersect'], [xIntersect],[yIntersect]):
        plt.annotate(
            label,
            xy=(x, y), xytext=(-20, 20),
            textcoords='offset points', ha='right', va='bottom',
            bbox=dict(boxstyle='round,pad=0.5', fc='yellow', alpha=0.5),
            arrowprops=dict(arrowstyle = '->', connectionstyle='arc3,rad=0'))

    plt.show()





    # leftRange_x=[leftRange[0]/srate,leftRange[1]/srate]
    #
    # plt.scatter(leftRange_x,[0,0])





    # plt.show()
    hh=1

def viz_blink_top_buttom_point(candidateSignal,blinkRange,blinkTop,blinkBottom,maxFrame,rightZero,leftRange,rightRange):
    # blinkTopPoint_ = np.argmax(candidateSignal[blinkRange] < blinkTop)
    #
    # blinkBottomPoint_ = np.argmin(candidateSignal[blinkRange] > blinkBottom) - 1  # The minus 1 is introduced in Python
    srate=1
    npad=20
    idx_t=(np.arange ( maxFrame-npad, rightZero +npad))
    bTrace= candidateSignal[idx_t]
    t=idx_t/srate
    #
    plt.plot(t, bTrace)
    #
    maxFrame_Y=candidateSignal[maxFrame]
    rightZero_Y=candidateSignal[rightZero]
    ee=[(maxFrame)/srate,(rightZero)/srate]
    ddd=[maxFrame_Y,rightZero_Y]
    #
    plt.scatter(ee,ddd)



    leftRange_x=[leftRange[0]/srate,leftRange[1]/srate]

    plt.scatter(leftRange_x,[0,0])

    rightRange_x=[rightRange[0]/srate,rightRange[1]/srate]

    plt.scatter(rightRange_x,[0,0])

    plt.show()
    hh=1


def cross_check_maxFrame(candidateSignal,maxFrame,srate):

    idx_t=(np.arange ( maxFrame-10, maxFrame +10))
    bTrace= candidateSignal[idx_t]
    t=idx_t/srate

    plt.plot(t, bTrace)

    ddd=candidateSignal[maxFrame-1]
    ee=(maxFrame-1)/srate

    plt.scatter(ee,ddd)

    plt.show()

def _viz_sanity_zero_crossing(start_f,stop_f,maxFrame,dsignal,dlegend=''):

    srate=1
    npad=20
    preLimit=start_f-npad
    postLimit=stop_f+npad
    idx_t=np.arange ( preLimit, postLimit)

    t=idx_t/srate
    bTrace =dsignal [idx_t]
    blinkScale = np.max(dsignal [np.arange ( preLimit-1, postLimit )])
    bTrace = bTrace*blinkScale/max(bTrace)
    plt.plot(t, bTrace)
    plt.scatter([start_f / srate, stop_f / srate],
                               [dsignal[start_f], dsignal[stop_f]],
                               marker='*',s=300)

    plt.scatter([maxFrame],
                [dsignal[maxFrame]],
                marker='o',s=300)

    plt.plot([t[0], t[-1]], [0, 0], 'r-', lw=2)
    plt.savefig('zero_Crossing_new_01.png')




import numpy as np
import mne
import os
import re
import shutil
import pandas as pd
def mad_matlab(arr, axis=None, keepdims=True):
    median = np.median(arr, axis=axis, keepdims=True)
    mad = np.median(np.abs(arr - median), axis=axis, keepdims=keepdims)[0]
    return mad

def check_make_folder(path, remove=False):
    if not remove:
        if not os.path.exists(path):
            os.makedirs(path, exist_ok=True)
    else:
        '''
        Some time I want the folder to be emptied
        '''
        if os.path.exists(path):
            shutil.rmtree(path)
            os.makedirs(path, exist_ok=True)
        else:
            os.makedirs(path, exist_ok=True)


def create_annotation(sblink, sfreq, label):
    st_blink = 'leftZero'
    blink_en = 'rightZero'
    # st_blink='startBlinks'
    # blink_en='endBlinks'
    if not isinstance(sblink, pd.DataFrame):
        raise ValueError('No appropriate channel. sorry. Try to use large channel selection')

    # d_s = ((sblink[blink_en] - sblink[st_blink]) / sfreq).tolist()
    # d_s = ((sblink[blink_en] - sblink[st_blink]) / sfreq).tolist()
    # onset_s = (sblink[st_blink] / sfreq).tolist()
    # d_s = ((sblink[blink_en] - sblink[st_blink]) / sfreq).tolist()
    onset_s = (sblink['maxFrames'] / sfreq).tolist()
    des_s = [label] * len(onset_s)
    d_s=[0]* len(onset_s)

    annot = mne.Annotations(onset=onset_s,  # in seconds
                            duration=d_s,  # in seconds, too
                            description=des_s)

    return annot


import logging

import pandas as pd

from pyblinkers import default_setting
from pyblinkers.vislab.blink_posi_vislab import BlinkFrameValley
from pyblinkers.utilities.extractBlinkProperties import BlinkProperties
from pyblinkers.utilities.fit_blink import FitBlinks
from pyblinkers.utilities.getCandidateSignal import PrepareSelection,ChannelSelection
from pyblinkers.utilities.misc import create_annotation
from pyblinkers.viz.viz_pd import viz_complete_blink_prop

logging.getLogger().setLevel(logging.INFO)


class BlinkDetector:
    def __init__(self, raw_data, visualize=False, annot_label=None,filter_bad=False):
        self.filter_bad=filter_bad
        self.raw_data = raw_data
        self.viz_data = visualize
        self.annot_label = annot_label
        self.sfreq = self.raw_data.info['sfreq']
        self.params = default_setting.params
        self.channel_list = self.raw_data.ch_names
        self.all_data_info = []
        self.all_data = []

    def prepare_raw_signal(self):
        self.raw_data.pick_types(eeg=True)
        self.raw_data.filter(0.5, 20.5, fir_design='firwin')
        self.raw_data.resample(100)
        return self.raw_data

    def process_channel_data(self, channel):

        df = BlinkFrameValley(self.raw_data.get_data(picks=channel)[0], self.sfreq, self.params, channel, self.sfreq).blink_frame
        df = FitBlinks(data=self.raw_data.get_data(picks=channel)[0], df=df).frame_blinks
        # df = extracBlinkProperties(self.raw_data.get_data(picks=channel)[0], df, self.sfreq)
        df=BlinkProperties(self.raw_data.get_data(picks=channel)[0], df, self.sfreq).df_res
        d = PrepareSelection(signal=self.raw_data.get_data(picks=channel)[0], df=df, params=self.params, ch=channel).get_param_for_selection()
        self.all_data_info.append(dict(df=df, ch=channel))
        self.all_data.append(d)

    @staticmethod
    def filter_point(ch,all_data_info):
        return list(filter(lambda all_data_info: all_data_info['ch'] == ch, all_data_info))[0]


    def filter_bad_blink(self,df):
        # filter_bad = False
        if self.filter_bad:
            df = df[df['blink_quality'] == 'Good']
        return df


    def generate_viz(self,data,df):
        fig_data = [viz_complete_blink_prop(data, row, self.sfreq) for index, row in df.iterrows()]

        return fig_data
    def get_blink_stat(self):
        for channel in self.channel_list:
            self.process_channel_data(channel)

        ch_blink_stat = pd.DataFrame(self.all_data)
        ch_selected = ChannelSelection(df=ch_blink_stat, params=self.params).df
        ch = ch_selected.loc[0, 'ch']
        nGoodBlinks = ch_selected.loc[0, 'numberGoodBlinks']
        data = self.raw_data.get_data(picks=ch)[0]
        rep_blink_channel = self.filter_point(ch,self.all_data_info)
        df = rep_blink_channel['df']

        df=self.filter_bad_blink(df)
        # df.to_pickle('unit_test_1.pkl')

        annot_description = self.annot_label if self.annot_label else 'eye_blink'
        annot = create_annotation(df, self.sfreq, annot_description)
        if self.viz_data:
            fig_data=self.generate_viz(data,df)
            return annot, ch, nGoodBlinks, fig_data, df

        return annot, ch, nGoodBlinks,df


import logging

import numpy as np
from tqdm import tqdm

from pyblinkers.utilities.misc import mad_matlab

logging.getLogger().setLevel(logging.INFO)


def getBlinkPosition(params, sfreq, blinkComp=None, ch='No_channel'):
    """

    % Parameters:
    %    blinkComp       independent component (IC) of eye-related
    %                    activations derived from EEG.  This component should
    %                    be "upright"
    %    srate:         sample rate at which the EEG data was taken
    %    stdTreshold    number of standard deviations above threshold for blink
    %    blinkPositions (output) 2 x n array with start and end frames of blinks
        :param params:
        :param sfreq:
        :param signal_eeg:
        :param ch:
        :return:
    """


    # Ensure 1D array
    assert blinkComp.ndim == 1
    # blinkComp = signal_eeg
    mu = np.mean(blinkComp, dtype=np.float64)

    mad_val = mad_matlab(blinkComp)
    robustStdDev = 1.4826 * mad_val

    minBlink = params['minEventLen'] * sfreq  # minimum blink frames
    threshold = mu + params['stdThreshold'] * robustStdDev  # actual threshold

    '''
    % The return structure.  Initially there is room for an event at every time
    % tick, to be trimmed later
    '''

    inBlink = False
    startBlinks = []
    endBlinks = []
    # kk=blinkComp.size
    for index in tqdm(range(blinkComp.size), desc=f"Get blink start and end for channel {ch}"):
        Drule = ~inBlink and (blinkComp[index] > threshold)
        if Drule:
            start = index
            inBlink = np.ones((1), dtype=bool)

        # if previous point was in a blink and signal retreats below threshold and duration greater than discard
        # threshold
        krule = (inBlink == True) and (blinkComp[index] < threshold)
        if krule:
            if (index - start) > minBlink:
                startBlinks.append(start)  # t_up
                endBlinks.append(index)  # t_dn

            inBlink = False

    arr_startBlinks = np.array(startBlinks)
    arr_endBlinks = np.array(endBlinks)

    # Now remove blinks that aren't separated
    positionMask = np.ones(arr_endBlinks.size, dtype=bool)

    x = (arr_startBlinks[1:] - arr_endBlinks[:-1]) / sfreq  # Calculate the blink duration
    y = np.argwhere(x <= 0.05)  # Index where blink duration is less than 0.05 sec
    positionMask[y] = np.zeros((1), dtype=bool)
    positionMask[y + 1] = np.zeros((1), dtype=bool)
    # v=arr_startBlinks[positionMask]
    blink_position = {'start_blink': arr_startBlinks[positionMask],
                      'end_blink': arr_endBlinks[positionMask],
                      'ch': ch}

    return blink_position



import numpy as np

params = {'stdThreshold': 1.50,
          'minEventLen': 0.05,
          'minEventSep': 0.05,
          'correlationThresholdTop': 0.980,
          'correlationThresholdBottom': 0.90,
          'params_blinkAmpRange_1': 0,
          'params_blinkAmpRange_2': 4,
          'params_goodRatioThreshold': 0.7,
          'params_minGoodBlinks': 10,
          'params_keepSignals': 0,
          'correlationThreshold': 0.98}



# def getHeader ():
#     header = ['mean', 'median', 'std',
#               'mad', 'goodMean', 'goodMedian',
#               'goodStd', 'goodMad']
#     return dict ( zip ( header, [np.nan] * len ( header ) ) )
#
#
# def getStatisticsStructure ():
#     list_detail = ['subjectID', 'task', 'uniqueName',
#                    'srate', 'startTime', 'usedNumber',
#                    'usedLabel', 'status', 'seconds',
#                    'numberBlinks','numberGoodBlinks',
#                    'goodRatio', 'header', 'pAVRZ',
#                    'nAVRZ', 'durationZ', 'durationB',
#                    'durationT', 'durationHZ',
#                    'durationHB', 'blinksPerMin']
#     return dict ( zip ( list_detail, [np.nan] * len ( list_detail ) ) )


import numpy as np

from pyblinkers.utilities.zero_crossing import (_maxPosVelFrame, _get_left_base, _get_right_base)


def create_left_right_base_vislab(data,df):
    blinkVelocity = np.diff(data, axis=0)
    df = df.dropna()
    df[['maxPosVelFrame', 'maxNegVelFrame']] = df.apply(
        lambda x: _maxPosVelFrame(blinkVelocity, x['maxFrames'], x['leftZero'],
                                  x['rightZero']), axis=1, result_type="expand")

    ## Lets check some condition especially for data with anamoly

    df = df[df['outerStarts'] < df['maxPosVelFrame']]  # Filter and take only row that normal
    df['leftBase'] = df.apply(lambda x: _get_left_base(blinkVelocity, x['outerStarts'], x['maxPosVelFrame']), axis=1)

    df = df.dropna()

    df['rightBase'] = df.apply(lambda x: _get_right_base(data, blinkVelocity, x['outerEnds'], x['maxNegVelFrame']),
                               axis=1)
    return df


import logging

import numpy as np
import pandas as pd

from pyblinkers.utilities.misc import mad_matlab

logging.getLogger().setLevel(logging.INFO)


def _goodblink_based_corr_median_std(df, correlationThreshold):
    R2 = df['leftR2'] >= correlationThreshold
    R3 = df['rightR2'] >= correlationThreshold

    # Now calculate the cutoff ratios -- use default for the values
    good_data = df.loc[R2.values & R3.values, :]
    bestValues = good_data['maxValues'].array

    specifiedMedian = np.nanmedian(bestValues)
    specifiedStd = 1.4826 * mad_matlab(bestValues)

    return R2, R3, specifiedMedian, specifiedStd


def get_mask_optimise(df, indicesNaN, correlationThreshold, zScoreThreshold):
    """
    "used Feb 02 2023"
    The calculation of bestmedian,worst median, worrst rbobustst
    is from https://github.com/VisLab/EEG-Blinks/blob/16b6ea04101ecfa74fb1c9cbceb037324572687e/blinker/utilities/extractBlinks.m#L97

    :param df:
    :param indicesNaN:
    :param correlationThreshold:
    :param zScoreThreshold:
    :return:
    """
    R1 = ~indicesNaN
    R2, R3, specifiedMedian, specifiedStd = _goodblink_based_corr_median_std(df, correlationThreshold)

    R4 = df['maxValues'] >= max(0, specifiedMedian - zScoreThreshold * specifiedStd)
    R5 = df['maxValues'] <= specifiedMedian + zScoreThreshold * specifiedStd
    bool_test = R1.values & R2.values & R3.values & R4.values & R5.values

    return bool_test, specifiedMedian, specifiedStd


def getGoodBlinkMask(df, zThresholds):
    "used Feb 02 2023"
    ## These is the default value
    correlationThreshold_s1, correlationThreshold_s2, zScoreThreshold_s1, zScoreThreshold_s2 = zThresholds

    df['rightR2'] = df['rightR2'].abs()
    df_data = df[['leftR2', 'rightR2', 'maxValues']]
    G=df_data.isnull()
    indicesNaN = df_data.isnull().any(axis='columns')

    ### GET MASK OPTIMISE

    goodMaskTop_bool, bestMedian, bestRobustStd = get_mask_optimise(df_data, indicesNaN, correlationThreshold_s1,
                                                                    zScoreThreshold_s1)

    df_s2_bool, worstMedian, worstRobustStd = get_mask_optimise(df_data, indicesNaN, correlationThreshold_s2,
                                                                zScoreThreshold_s2)

    goodBlinkMask = np.reshape(goodMaskTop_bool | df_s2_bool, (-1, 1))  # Get any TRUE
    df['blink_quality'] = 'Good'
    df[['blink_quality']] = df[['blink_quality']].where(goodBlinkMask, other='Reject')
    return df, bestMedian, bestRobustStd




class BlinkProperties:
    '''
    Return a structure with blink shapes and properties for individual blinks
    '''
    def __init__(self, data, df, srate):
        self.data = data
        self.df = df
        self.srate = srate
        self.shutAmpFraction = 0.9
        self.pAVRThreshold = 3
        self.zThresholds = (0.90, 0.98, 2, 5)
        self.df_res=[]
        self.reset_index()
        self.get_good_blink_mask()
        self.set_blink_velocity()
        self.set_blink_duration()
        self.set_blink_amp_velocity_ratio_zero_to_max()
        self.amplitude_velocity_ratio_base()
        self.amplitude_velocity_ratio_tent()
        self.time_zero_shut()
        self.time_base_shut()
        self.extract_other_times()

    def reset_index(self):
        self.df.reset_index(drop=True, inplace=True)

    def get_good_blink_mask(self):
        self.df, self.bestMedian, self.bestRobustStd = getGoodBlinkMask(self.df, self.zThresholds)

    def set_blink_velocity(self):
        self.signal_l = self.data.shape[0]
        self.blinkVelocity = np.diff(self.data)

    def set_blink_duration(self):
        cols_int = ['rightBase']
        self.df[cols_int] = self.df[cols_int].astype(int)
        self.df['durationBase'] = (self.df['rightBase'] - self.df['leftBase']) / self.srate
        self.df['durationTent'] = (self.df['rightXIntercept'] - self.df['leftXIntercept']) / self.srate
        self.df['durationZero'] = (self.df['rightZero'] - self.df['leftZero']) / self.srate
        self.df['durationHalfBase'] = (self.df['rightBaseHalfHeight'] - self.df['leftBaseHalfHeight'] + 1) / self.srate
        self.df['durationHalfZero'] = (self.df['rightZeroHalfHeight'] - self.df['leftZeroHalfHeight'] + 1) / self.srate

    def set_blink_amp_velocity_ratio_zero_to_max(self):
        self.df[['leftZero', 'rightZero']] = self.df[['leftZero', 'rightZero']].astype(int)
        self.df['peaksPosVelZero'] = self.df.apply(lambda x: x['leftZero'] + np.argmax(self.blinkVelocity[x['leftZero']:x['maxFrames'] + 1]), axis=1)
        self.df['RRC'] = self.data[self.df['maxFrames'] - 1] / self.blinkVelocity[self.df['peaksPosVelZero']]
        self.df['posAmpVelRatioZero'] = (100 * abs(self.df['RRC'])) / self.srate
        self.df['downStrokevelFrame_del'] = self.df.apply(lambda x: x['maxFrames'] + np.argmin(self.blinkVelocity[x['maxFrames']:x['rightZero'] + 1]), axis=1)
        self.df['TTT'] = self.data[self.df['maxFrames'] - 1] / self.blinkVelocity[self.df['downStrokevelFrame_del']]
        self.df['negAmpVelRatioZero'] = (100 * abs(self.df['TTT'])) / self.srate

    def amplitude_velocity_ratio_base(self):
        self.df['peaksPosVelBase'] = self.df.apply(
            lambda x: x['leftBase'] + np.argmax(self.blinkVelocity[x['leftBase']:x['maxFrames'] + 1]), axis=1)
        self.df['KKK'] = self.data[self.df['maxFrames'] - 1] / self.blinkVelocity[self.df['peaksPosVelBase']]
        self.df['posAmpVelRatioBase'] = (100 * abs(self.df['KKK'])) / self.srate

        self.df['downStroke_del'] = self.df.apply(
            lambda x: x['maxFrames'] + np.argmin(self.blinkVelocity[x['maxFrames']:x['rightBase'] + 1]), axis=1)
        self.df['KKK'] = self.data[self.df['maxFrames'] - 1] / self.blinkVelocity[self.df['downStroke_del']]
        self.df['negAmpVelRatioBase'] = (100 * abs(self.df['KKK'])) / self.srate

    def get_argmax_val(self,row):
        left = row['leftXIntercept_int']
        right = row['rightXIntercept_int'] + 1
        start = row['start_shut_tst']
        max_val = row['maxValues']
        shut_amp_frac = self.shutAmpFraction

        subset = self.data[left:right][start:-1]
        dconstant=shut_amp_frac * max_val

        try:
            return np.argmax(subset<dconstant)
        except ValueError:
            return np.nan
    def amplitude_velocity_ratio_tent(self):
        self.df['pop'] = self.data[self.df['maxFrames'] - 1] / self.df['averRightVelocity']
        self.df['negAmpVelRatioTent'] = (100 * abs(self.df['pop'])) / self.srate

        self.df['opi'] = self.data[self.df['maxFrames'] - 1] / self.df['averLeftVelocity']
        self.df['WE'] = (100 * abs(self.df['opi']))
        self.df['posAmpVelRatioTent'] = self.df['WE'] / self.srate

    def time_zero_shut(self):
        self.df['closingTimeZero'] = (self.df['maxFrames'] - self.df['leftZero']) / self.srate
        self.df['reopeningTimeZero'] = (self.df['rightZero'] - self.df['maxFrames']) / self.srate

        self.df['ampThreshhold'] = self.shutAmpFraction * self.df['maxValues']
        self.df['start_shut_tzs'] = self.df.apply(
            lambda x: np.argmax(self.data[x['leftZero']:x['rightZero'] + 1] >= x['ampThreshhold']), axis=1)

        self.df['endShut_tzs'] = self.df.apply(
            lambda x: np.argmax(self.data[x['leftZero']:x['rightZero'] + 1][x['start_shut_tzs'] + 1:-1] <
                                self.shutAmpFraction * x['maxValues']), axis=1)

        ## PLease expect error here, some value maybe zero or lead to empty cell
        self.df['endShut_tzs'] = self.df['endShut_tzs'] + 1  ## temporary  to delete
        self.df['timeShutZero'] = self.df.apply(
            lambda x: 0 if x['endShut_tzs'] == np.isnan else x['endShut_tzs'] / self.srate, axis=1)
    def time_base_shut(self):
        self.df['ampThreshhold_tbs'] = self.shutAmpFraction * self.df['maxValues']
        self.df['start_shut_tbs'] = self.df.apply(
            lambda x: np.argmax(self.data[x['leftBase']:x['rightBase'] + 1] >= x['ampThreshhold_tbs']), axis=1)

        self.df['endShut_tbs'] = self.df.apply(
            lambda x: np.argmax(self.data[x['leftBase']:x['rightBase'] + 1][x['start_shut_tbs']:-1] <
                                self.shutAmpFraction * x['maxValues']), axis=1)

        self.df['timeShutBase'] = self.df.apply(
            lambda x: 0 if x['endShut_tbs'] == np.isnan else (x['endShut_tbs'] / self.srate), axis=1)

        ## Time shut tent
        self.df['closingTimeTent'] = (self.df['xIntersect'] - self.df['leftXIntercept']) / self.srate
        self.df['reopeningTimeTent'] = (self.df['rightXIntercept'] - self.df['xIntersect']) / self.srate

        self.df['ampThreshhold_tst'] = self.shutAmpFraction * self.df['maxValues']

        self.df[['leftXIntercept_int', 'rightXIntercept_int']] = self.df[['leftXIntercept', 'rightXIntercept']].astype(
            int)


        # warnings.warn('New upgrade')
        self.df=self.df[self.df.leftXIntercept_int<self.df.rightXIntercept_int]
        self.df.reset_index(drop=True,inplace=True)
        self.df['start_shut_tst'] = self.df.apply(
            lambda x: np.argmax(self.data[x['leftXIntercept_int']:x['rightXIntercept_int'] + 1] >= x['ampThreshhold']), axis=1)


        self.df['endShut_tst'] = self.df.apply(self.get_argmax_val,axis=1)

        ### Just in case got other issue, use this back, do not delete
        ## Since we already remove imblanace left_right intercept, maybe dah xperlu kot bawah ni
        # df['start_shut_tst'] = df.apply(
        #     lambda x: _start_shut(data, x['leftXIntercept_int'], x['rightXIntercept_int'], x['ampThreshhold']), axis=1)
        #
        #
        #
        # df['endShut_tst'] = df.apply(
        #     lambda x: _end_shut(data, x['leftXIntercept_int'], x['rightXIntercept_int'], x['start_shut_tst'],
        #                         x['maxValues'], shutAmpFraction), axis=1)
        vv=1
        self.df['timeShutTent'] = self.df.apply(
            lambda x: 0 if x['endShut_tst'] == np.isnan else (x['endShut_tst'] / self.srate), axis=1)

    def extract_other_times(self):
        ## Other times
        self.df['peakMaxBlink '] = self.df['maxValues']
        self.df['peakMaxTent'] = self.df['yIntersect']
        self.df['peakTimeTent'] = self.df['xIntersect'] / self.srate
        self.df['peakTimeBlink'] = self.df['maxFrames'] / self.srate

        dfcal = self.df[['maxFrames', 'peaksPosVelBase', 'peaksPosVelZero']]

        df_t = pd.DataFrame.from_records([[self.signal_l] * 3], columns=['maxFrames', 'peaksPosVelBase', 'peaksPosVelZero'])

        dfcal = pd.concat([dfcal, df_t]).reset_index(drop=True)

        dfcal['ibmx'] = dfcal.maxFrames.diff().shift(-1)

        dfcal['interBlinkMaxAmp'] = dfcal['ibmx'] / self.srate

        dfcal['ibmvb'] = 1 - dfcal['peaksPosVelBase']
        dfcal['interBlinkMaxVelBase'] = dfcal['ibmvb'] / self.srate  # peaksPosVelBase == velFrame

        dfcal['ibmvz'] = 1 - dfcal['peaksPosVelZero']
        dfcal['interBlinkMaxVelZero'] = dfcal['ibmvz'] / self.srate

        dfcal.drop(dfcal.tail(1).index, inplace=True)  # drop last n rows# peaksPosVelZero == velFrame
        dfnew = self.df[['maxValues', 'posAmpVelRatioZero']]

        R1 = dfnew['posAmpVelRatioZero'] < self.pAVRThreshold

        th_bm_brs = self.bestMedian - self.bestRobustStd
        R2 = dfnew['maxValues'] < th_bm_brs
        pMask = pd.concat([R1, R2], axis=1)
        pMasks = pMask.all(1)
        df_res = pd.merge(self.df, dfcal, on=['maxFrames'])
        self.df_res = df_res[~pMasks].reset_index(drop=True)
        # return df_res




from setuptools import setup

setup(name='pyblinkers',
      version='0.0.6',
      description='make life easier',
      author='rpb',
      packages=['pyblinkers','pyblinkers.utilities',
                  'pyblinkers.vislab','pyblinkers.viz'],
      long_description='Get EOG from EEG signal recording. In this 0.0.6 version, we change the name from eeg-blinks to pyblinker',
      install_requires=['seaborn']
      )



import logging
import numpy as np
import pandas as pd
from pyblinkers.vislab.getBlinkPositions_vislab import getBlinkPosition
from pyblinkers.utilities.zero_crossing import left_right_zero_crossing

logging.basicConfig(level=logging.INFO)

class BlinkFrameValley:
    def __init__(self, data, sfreq, params, ch, srate):
        self.data = data
        self.sfreq = sfreq
        self.params = params
        self.ch = ch
        self.srate = srate
        self._get_blink_position()
        self._get_max_values()
        self._compute_duration()
        self._filter_duration()
        self._get_zero_crossing()

    def _get_blink_position(self):
        self.ch_data = getBlinkPosition(self.params, self.sfreq, blinkComp=self.data, ch=self.ch)
        self.startBlinks = self.ch_data['start_blink']
        self.endBlinks = self.ch_data['end_blink']

    def _get_max_frame(self, startBlinks, endBlinks):
        blinkRange = np.arange(startBlinks, endBlinks + 1)
        blink_frame = self.data[startBlinks:endBlinks + 1]
        maxValues = np.amax(blink_frame)
        maxFrames = blinkRange[np.argmax(blink_frame)]
        return maxValues, maxFrames

    def _get_max_values(self):
        maxValues, maxFrames = zip(*[self._get_max_frame(dstartBlinks, dendBlinks) for
                                     dstartBlinks, dendBlinks in zip(self.startBlinks, self.endBlinks)])
        self.maxFrames = np.array(maxFrames)
        self.maxValues = np.array(maxValues)
        self.outerStarts = np.append(0, self.maxFrames[0:-1])
        self.outerEnds = np.append(self.maxFrames[1:], self.data.size)

    def _compute_duration(self):
        self.blink_frame = pd.DataFrame(dict(maxFrames=self.maxFrames, maxValues=self.maxValues,
                                             startBlinks=self.startBlinks, endBlinks=self.endBlinks,
                                             outerStarts=self.outerStarts, outerEnds=self.outerEnds))
        self.blink_frame['blink_duration'] = (self.blink_frame['endBlinks'] - self.blink_frame['startBlinks']) / self.srate

    def _filter_duration(self):
        self.blink_frame = self.blink_frame[self.blink_frame.blink_duration.ge(0.05)].reset_index(drop=True)



    def _get_zero_crossing(self):
        self.blink_frame[['leftZero', 'rightZero']] = self.blink_frame.apply(lambda x: left_right_zero_crossing(self.data,x['maxFrames'], x['outerStarts'],
                                                                                        x['outerEnds']), axis=1,
                                                     result_type="expand")







import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns

sns.set_style("darkgrid")

def viz_complete_blink_prop(data,row,srate):


    """

    TODO Viz

    https://stackoverflow.com/a/51928241/6446053
    https://stackoverflow.com/a/38015084/6446053
    :return:
    """

    bquality = row['blink_quality']
    blinkq = 'Gblink' if bquality else 'Bblink'
    xLabelString = f"Time(s) [Max:F={row['maxFrames']},T={(row['maxFrames'])/srate}] " \
                   f"[Left:F={row['leftBase']},T={(row['rightBase'])/srate} sec].blinkQuality:{blinkq}"

    fig, ax = plt.subplots(figsize=(8, 6))

    npad = 20
    preLimit = row['startBlinks'] - npad
    postLimit = row['endBlinks'] + npad

    idx_t = np.arange(preLimit, postLimit + 1)

    bTrace = data[idx_t]


    plt.plot(idx_t, bTrace,linestyle='-',marker='o',color='b',
             label='line with marker',alpha=0.7)
    plt.plot([idx_t[0], idx_t[-1]], [0, 0], "--", color="gray", lw=2,label='Y0')


    plt.plot([row['xLineCross_l'] , row['xIntersect']], [row['yLineCross_l'],  row['yIntersect']], "--", color="gray", lw=2)
    plt.plot([row['xIntersect'],row['xLineCross_r']], [row['yIntersect'],row['yLineCross_r']], "--", color="gray", lw=2)

    ## PLot key point
    plt.scatter([row['blinkBottomPoint_l_X'],row['blinkTopPoint_l_X']],
                [row['blinkBottomPoint_l_Y'],row['blinkTopPoint_l_Y']],
                marker='*', s=200,label='left_top_down_blink')

    plt.scatter([row['blinkBottomPoint_r_X'],row['blinkTopPoint_r_X']],
                [row['blinkBottomPoint_r_Y'],row['blinkTopPoint_r_Y']],
                marker='*', s=200,label='right_top_down_blink')



    plt.scatter(row['xIntersect'], row['yIntersect'],label='tent_point')


    plt.scatter([row['leftZero'], row['rightZero']], [0, 0], marker='d', s=100,label='zero crossing')
    plt.scatter(row['maxFrames'], data[row['maxFrames']],label='max Frame')

    plt.scatter([row['leftBase'], row['rightBase']],
                [data[row['leftBase']], data[row['rightBase']]],label='base')

    plt.scatter([row['leftBaseHalfHeight'], row['rightBaseHalfHeight']],
                [data[row['leftBaseHalfHeight']], data[row['rightBaseHalfHeight']]],
                marker='<', s=200,label='BaseHalfHeight')

    plt.scatter([row['rightZeroHalfHeight'], row['leftZeroHalfHeight']],
                [data[row['rightZeroHalfHeight']], data[row['leftZeroHalfHeight']]],
                marker='>', s=300,label='ZeroHalfHeight')

    plt.scatter([row['maxPosVelFrame'],  row['maxNegVelFrame']],
                [data[row['maxPosVelFrame']], data[ row['maxNegVelFrame']]],
                marker='>', s=300,label='maxVelFrame')



    plt.legend()
    ylabel='Signal(uv)'
    plt.xlabel(xLabelString)
    plt.ylabel(ylabel)
    bquality=row['blink_quality'] == 'Good'
    maxFrames=row['maxFrames']
    d=dict(fig=fig,blink_quality=bquality,maxFrames=maxFrames)

    return d